#include "Arduino.h"
#include "config.h"
#include "def.h"
#include "types.h"
#include "EEPROM.h"
#include "LCD.h"
#include "Output.h"
#include "GPS.h"
#include "MultiWii.h"
#include "Serial.h"
#include "Protocol.h"
#include "RX.h"

#include "EQ.h"

#include "Command.h"
// -----------------------------------------------------------------------------------
// Command processing

static uint8_t CURRENTPORT = 0;

// scratch-pad variables
double f0, f1, f2, f3;
int    i, i1, i2;
byte   b;
unsigned long _coord_t = 0;
double _dec, _ra;

// help with commands
enum Command {COMMAND_NONE, COMMAND_SERIAL, COMMAND_SERIAL1, COMMAND_SERIAL4, COMMAND_SERIALX};
char reply[50];
char command[3];
char parameter[25];
boolean commandError = false;
boolean quietReply   = false;
cb cmd;  // the first Serial is always enabled
char replyx[50] = "";
cb cmdx; // virtual command channel for internal use


//#define DEBUG


// process commands
void processCommands()
{
  boolean supress_frame = false;
  char *conv_end;

  if ((SerialLX200Available(0) > 0) && (!cmd.ready()))
  {
    cmd.add(SerialLX200Read(0));
  }

  // if a command is ready, process it
  Command process_command = COMMAND_NONE;
  if (cmd.ready())
  {
#ifdef DEBUG
    SerialLX200String(0, "\r\nCmd:[");
#endif
    strcpy(command, cmd.getCmd());
#ifdef DEBUG
    SerialLX200String(0, command);
#endif
    strcpy(parameter, cmd.getParameter());
#ifdef DEBUG
    SerialLX200String(0, "] Parameter:[");
    SerialLX200String(0, parameter);
    SerialLX200String(0, "]\r\n");
#endif
    cmd.flush();
    process_command = COMMAND_SERIAL;
  }
  else if (cmdx.ready())
  {
    strcpy(command, cmdx.getCmd());
    strcpy(parameter, cmdx.getParameter());
    cmdx.flush();
    process_command = COMMAND_SERIALX;
  }
  else
    return;

  if (process_command) {
    // Command is two chars followed by an optional parameter...
    commandError = false;
    // Handles empty and one char replies
    reply[0] = 0;
    reply[1] = 0;

    //   (char)6 - Special
    if (command[0] == (char)6) {
      if (command[1] == '0') {
        reply[0] = command[1];
        reply[1] = command[1];
        reply[2] = 0; // last cmd checksum failed, 00#
      } else {
        reply[0] = command[1];
        reply[1] = 0; // Equatorial or Horizon mode, A or P
        // supress_frame = true;
      }
      // quietReply = true;
    }
    else if (command[0] == 'A') {
      //   A - Alignment Commands
      //  :AW#  Align Write to EEPROM
      //         Returns: 1 on success
      if (command[1] == 'W')
      {
      }
      else if (command[1] == '?')
      {
        //  :A?#  Align status
        //         Returns: mno#
        //         where m is the maximum number of alignment stars
        //               n is the current alignment star (0 otherwise)
        //               o is the last required alignment star when an alignment is in progress (0 otherwise)
      }
      else if ((command[1] >= '1') && (command[1] <= '9'))
      {
        //  :An#  Start Telescope Manual Alignment Sequence
        //         This is to initiate a one or two-star alignment:
        //         1) Before calling this function, the telescope should be in the polar-home position
        //         2) Call this function
        //         3) Set the target location (RA/Dec) to a bright star, etc. (near the celestial equator in the western sky)
        //         4) Issue a goto command
        //         5) Center the star/object using the guide commands (as needed)
        //         6) Call :A+# command to accept the correction
        //         ( for two-star alignment )
        //         7) Set the target location (RA/Dec) to a bright star, etc. (near the celestial equator in the southern sky)
        //         8) Issue a goto command
        //         9) Center the star/object using the guide commands (as needed)
        //         10) Call :A+# command to accept the correction
        //         Returns:
        //         1: When ready for your goto commands
        //         0: If mount is busy
      }
      else if (command[1] == '+')
      {
        //  :A+#  Manual Alignment, set target location
        //         Returns:
        //         1: If correction is accepted
        //         0: Failure, Manual align mode not set or distance too far
      }
    }
    else if ((command[0] == 'C') && ((command[1] == 'S') || command[1] == 'M'))  {
      //   C - Sync Control

      //  :CS#   Synchonize the telescope with the current right ascension and declination coordinates
      //         Returns: Nothing (Sync's fail silently)
      //  :CM#   Synchonize the telescope with the current database object (as above)
      //         Returns: "N/A#" on success, "En#" on failure where n is the error code per the :MS# command
      if ((parkStatus == NotParked) && (trackingState != TrackingMoveTo)) {
        i = syncEqu(newTargetRA, newTargetDec);
        if (command[1] == 'M') {
          if (i == 0) strcpy(reply, "N/A");
          if (i > 0) {
            reply[0] = 'E';
            reply[1] = '0' + i;
            reply[2] = 0;
          }
        }
        quietReply = true;
      }
    }
    else if (command[0] == 'G')
    {
      //   G - Get Telescope Information

      //  :GA#   Get Telescope Altitude
      //         Returns: sDD*MM# or sDD*MM'SS# (based on precision setting)
      //         The current scope altitude
      if (command[1] == 'A')
      {
        // getHor(&f0, &f1);
        if (!doubleToDms(reply, &f0, false, true))
          commandError = true;
        else
          quietReply = true;
      }
      else if (command[1] == 'a')
      {
        //  :Ga#   Get Local Time in 12 hour format
        //         Returns: HH:MM:SS#
        i = highPrecision;
        highPrecision = true;
        LMT = timeRange(UT1 - timeZone);
        if (LMT > 12.0) LMT -= 12.0;
        if (!doubleToHms(reply, &LMT)) commandError = true;
        else quietReply = true;
        highPrecision = i;
      }
      else if (command[1] == 'C') {
        //  :GC#   Get the current date
        //         Returns: MM/DD/YY#
        //         The current local calendar date
        LMT = UT1 - timeZone;
        // correct for day moving forward/backward... this works for multipule days of up-time
        double J = JD;
        int y, m, d;
        while (LMT >= 24.0) {
          LMT = LMT - 24.0;
          J = J - 1.0;
        }
        if    (LMT < 0.0)   {
          LMT = LMT + 24.0;
          J = J + 1.0;
        }
        greg(J, &y, &m, &d); y -= 2000; if (y >= 100) y -= 100;
        sprintf(reply, "%02d/%02d/%02d", m, d, y);
        quietReply = true;
      }
      else if (command[1] == 'c') {
        //  :Gc#   Get the current time format
        //         Returns: 24#
        //         The current local time format
        strcpy(reply, "24");
        quietReply = true;
      }
      else if (command[1] == 'D')  {
        //  :GD#   Get Telescope Declination
        //         Returns: sDD*MM# or sDD*MM'SS# (based on precision setting)
        // /*
#ifdef HAL_SLOW_PROCESSOR
        if (millis() - _coord_t < 100)
#else
        if (millis() - _coord_t < 5)
#endif
        {
          f0 = _ra;
          f1 = _dec;
        }
        else
        {
          getEqu(&f0, &f1, false);
          f0 /= 15.0;
          _ra = f0;
          _dec = f1;
          _coord_t = millis();
        }
        // */
        if (!doubleToDms(reply, &f1, false, true))
          commandError = true;
        else
          quietReply = true;
      }
      else if (command[1] == 'd')  {
        //  :Gd#   Get Currently Selected Target Declination
        //         Returns: sDD*MM# or sDD*MM'SS# (based on precision setting)
        if (!doubleToDms(reply, &newTargetDec, false, true)) commandError = true;
        else quietReply = true;
      }
      else if (command[1] == 'G')  {
        //  :GG#   Get UTC offset time
        //         Returns: sHH#
        //         The number of decimal hours to add to local time to convert it to UTC
        timeZoneToHM(reply, timeZone);
        quietReply = true;
      } else if (command[1] == 'g')  {
        //  :Gg#   Get Current Site Longitude
        //         Returns: sDDD*MM#
        //         The current site Longitude. East Longitudes are negative
        i = highPrecision;
        highPrecision = false;
        if (!doubleToDms(reply, &longitude, true, true)) commandError = true;
        else quietReply = true;
        highPrecision = i;
      } else if (command[1] == 'h')  {
        //  :Gh#   Get Horizon Limit
        //         Returns: sDD*#
        //         The minimum elevation of an object above the horizon required for a mount goto
        sprintf(reply, "%+02d*", minAlt);
        quietReply = true;
      } else if (command[1] == 'L')  {
        //  :GL#   Get Local Time in 24 hour format
        //         Returns: HH:MM:SS#
        //         On devices with single precision fp several days up-time will cause loss of precision as additional mantissa digits are needed to represent hours
        //         Devices with double precision fp are limitated by sidereal clock overflow which takes 249 days
        i = highPrecision;
        highPrecision = true;
        LMT = timeRange(UT1 - timeZone);
        if (!doubleToHms(reply, &LMT)) commandError = true;
        else quietReply = true;
        highPrecision = i;
      }
      else if ((command[1] == 'M') || (command[1] == 'N') || (command[1] == 'O') || (command[1] == 'P'))  {
        //  :GM#   Get Site 1 Name
        //  :GN#   Get Site 2 Name
        //  :GO#   Get Site 3 Name
        //  :GP#   Get Site 4 Name
        //         Returns: <string>#
        //         A # terminated string with the name of the requested site.
        i = command[1] - 'M';
        // EEPROM_readString(EE_sites + i * 25 + 9, reply);
        if (reply[0] == 0) {
          strcat(reply, "None");
        }
        quietReply = true;
      }
      else if (command[1] == 'm')  {
        //  :Gm#   Gets the meridian pier-side
        //         Returns: E#, W#, N# (none/parked), ?# (Meridian flip in progress)
        //         A # terminated string with the pier side.
        reply[0] = '?'; reply[1] = 0;
        if (pierSide == PierSideNone) reply[0] = 'N';
        if (pierSide == PierSideEast) reply[0] = 'E';
        if (pierSide == PierSideWest) reply[0] = 'W';
        quietReply = true;
      }
      else if (command[1] == 'o')  {
        //  :Go#   Get Overhead Limit
        //         Returns: DD*#
        //         The highest elevation above the horizon that the telescope will goto
        sprintf(reply, "%02d*", maxAlt);
        quietReply = true;
      }
      else if (command[1] == 'R')  {
        //  :GR#   Get Telescope RA
        //         Returns: HH:MM.T# or HH:MM:SS# (based on precision setting)
        // /*
#ifdef HAL_SLOW_PROCESSOR
        if (millis() - _coord_t < 100)
#else
        if (millis() - _coord_t < 5)
#endif
        {
          f0 = _ra;
          f1 = _dec;
        }
        else
        {
          getEqu(&f0, &f1, false);
          f0 /= 15.0;
          _ra = f0;
          _dec = f1;
          _coord_t = millis();
        }
        // */
        if (!doubleToHms(reply, &f0))
          commandError = true;
        else
          quietReply = true;
      }
      else if (command[1] == 'r')  {
        //  :Gr#   Get current/target object RA
        //         Returns: HH:MM.T# or HH:MM:SS (based on precision setting)
        f0 = newTargetRA;
        f0 /= 15.0;
        if (!doubleToHms(reply, &f0)) commandError = true;
        else quietReply = true;
      }
      else if (command[1] == 'S')  {
        //  :GS#   Get the Sidereal Time
        //         Returns: HH:MM:SS#
        //         The Sidereal Time as an ASCII Sexidecimal value in 24 hour format
        i = highPrecision;
        highPrecision = true;
        f0 = LST();
        if (!doubleToHms(reply, &f0)) commandError = true;
        else quietReply = true;
        highPrecision = i;
      }
      else if (command[1] == 'T')  {
        //  :GT#   Get tracking rate
        //         Returns: dd.ddddd# (OnStep returns more decimal places than LX200 standard)
        //         Returns the tracking rate if siderealTracking, 0.0 otherwise
        if (trackingState == TrackingSidereal) {
#ifdef MOUNT_TYPE_ALTAZM
          f0 = GetTrackingRate() * 1.00273790935 * 60.0;
#else
          f0 = (trackingTimerRateAxis1 * 1.00273790935) * 60.0;
#endif
        }
        else f0 = 0.0;
        char temp[10];
        dtostrf(f0, 0, 5, temp);
        strcpy(reply, temp);
        quietReply = true;
      }
      else if (command[1] == 't')  {
        //  :Gt#   Get Current Site Latitude
        //         Returns: sDD*MM#
        //         The latitude of the current site. Positive for North latitudes
        i = highPrecision;
        highPrecision = false;
        if (!doubleToDms(reply, &latitude, false, true)) commandError = true;
        else quietReply = true;
        highPrecision = i;
      }
      else if (command[1] == 'U')  {
        //  :GU#   Get telescope Status
        //         Returns: SS#
        i = 0;
        if (trackingState != TrackingSidereal)     reply[i++] = 'n';
        if (trackingState != TrackingMoveTo)       reply[i++] = 'N';
        const char *parkStatusCh = "pIPF";       reply[i++] = parkStatusCh[parkStatus]; // not [p]arked, parking [I]n-progress, [P]arked, Park [F]ailed
        if (pecRecorded)                         reply[i++] = 'R';
        if (pecAutoRecord)                       reply[i++] = 'W';
        if (atHome)                              reply[i++] = 'H';
        if (PPSsynced)                           reply[i++] = 'S';
        if ((guideDirAxis1) || (guideDirAxis2))  reply[i++] = 'G';
        if (faultAxis1 || faultAxis2)            reply[i++] = 'f';
        if (refraction)                          reply[i++] = 'r'; else reply[i++] = 's';
        if (onTrack)                             reply[i++] = 't';
        if (waitingHome)                         reply[i++] = 'w';
        // provide mount type
#ifdef MOUNT_TYPE_GEM
        reply[i++] = 'E';
#endif
#ifdef MOUNT_TYPE_FORK
        reply[i++] = 'K';
#endif
#ifdef MOUNT_TYPE_FORK_ALT
        reply[i++] = 'k';
#endif
#ifdef MOUNT_TYPE_ALTAZM
        reply[i++] = 'A';
#endif
        reply[i++] = '0' + lastError;
        reply[i++] = 0;
        quietReply = true;
      }
      else if (command[1] == 'V') {
        //  :GVD# Get Telescope Firmware Date
        //         Returns: mmm dd yyyy#
        //  :GVN# Get Telescope Firmware Number
        //         Returns: d.dc#
        //  :GVP# Get Telescope Product Name
        //         Returns: <string>#
        //  :GVT# Get Telescope Firmware Time
        //         returns: HH:MM:SS#
        if (parameter[1] == (char)0) {
          if (parameter[0] == 'D')
            // strcpy(reply, FirmwareDate)
            ;
          else if (parameter[0] == 'N')
            // sprintf(reply, "%i.%i%s", FirmwareVersionMajor, FirmwareVersionMinor, FirmwareVersionPatch)
            ;
          else if (parameter[0] == 'P')
            // strcpy(reply, FirmwareName)
            ;
          else if (parameter[0] == 'T')
            // strcpy(reply, FirmwareTime)
            ;
          else
            commandError = true;
        } else
          commandError = true;

        quietReply = true;
      }
      else if (command[1] == 'X')  {
        //  :GXnn#   Get OnStep value
        //         Returns: value
        if (parameter[2] == (char)0) {
          if (parameter[0] == '0') { // 0n: Align Model
            switch (parameter[1]) {
              case '0':
                sprintf(reply, "%ld", (long)(indexAxis1 * 3600.0));
                quietReply = true; break; // indexAxis1
              case '1':
                sprintf(reply, "%ld", (long)(indexAxis2 * 3600.0));
                quietReply = true; break; // indexAxis2
              case '2':
                // sprintf(reply, "%ld", (long)(GeoAlign.altCor * 3600.0));
                quietReply = true; break; // altCor
              case '3':
                // sprintf(reply, "%ld", (long)(GeoAlign.azmCor * 3600.0));
                quietReply = true; break; // azmCor
              case '4':
                // sprintf(reply, "%ld", (long)(GeoAlign.doCor * 3600.0));
                quietReply = true; break; // doCor
              case '5':
                // sprintf(reply, "%ld", (long)(GeoAlign.pdCor * 3600.0));
                quietReply = true; break; // pdCor
#if defined(MOUNT_TYPE_FORK) || defined(MOUNT_TYPE_FORK_ALT)
              case '6':
                sprintf(reply, "%ld", (long)(GeoAlign.dfCor * 3600.0));
                quietReply = true; break; // ffCor
              case '7':
                sprintf(reply, "%ld", (long)(0));
                quietReply = true; break;                  // dfCor
#else
              case '6':
                sprintf(reply, "%ld", (long)(0));
                quietReply = true; break;                  // ffCor
              case '7':
                // sprintf(reply, "%ld", (long)(GeoAlign.dfCor * 3600.0));
                quietReply = true; break; // dfCor
#endif
              case '8':
                // sprintf(reply, "%ld", (long)(GeoAlign.tfCor * 3600.0));
                quietReply = true; break; // tfCor
              default:
                commandError = true;
            }
          }
          else if (parameter[0] == '8')
          { // 8n: Date/Time
            switch (parameter[1]) {
              case '0':
                i = highPrecision;
                highPrecision = true;
                f0 = timeRange(UT1);
                doubleToHms(reply, &f0);
                highPrecision = i;
                quietReply = true;
                break; // UTC time
              case '1': f1 = JD; f0 = UT1; while (f0 >= 24.0) {
                  f0 -= 24.0;
                  f1 += 1;
                } while (f0 < 0.0) {
                  f0 += 24.0;
                  f1 -= 1;
                } greg(f1, &i2, &i, &i1); i2 = (i2 / 99.99999 - floor(i2 / 99.99999)) * 100; sprintf(reply, "%02d/%02d/%02d", i, i1, i2); quietReply = true; break; // UTC date
              default:  commandError = true;
            }
          }
          else if (parameter[0] == '9') { // 9n: Misc.
            switch (parameter[1]) {
              case '0': dtostrf(guideRates[currentPulseGuideRate] / 15.0, 2, 2, reply); quietReply = true; break; // pulse-guide rate
              case '1': sprintf(reply, "%i", pecAnalogValue); quietReply = true; break;                     // pec analog value
              case '2': sprintf(reply, "%ld", (long)(maxRate / 16L)); quietReply = true; break;             // MaxRate
              case '3': sprintf(reply, "%ld", (long)(MaxRate)); quietReply = true; break;                   // MaxRate (default)
              case '4': if (meridianFlip == MeridianFlipNever) {
                  sprintf(reply, "%d N", (int)(pierSide));
                } else {
                  sprintf(reply, "%d", (int)(pierSide));
                } quietReply = true; break; // pierSide (N if never)
              case '5': sprintf(reply, "%i", (int)autoMeridianFlip); quietReply = true; break;              // autoMeridianFlip
              case '6':                                                                                     // preferred pier side
                if (preferredPierSide == PPS_EAST) strcpy(reply, "E"); else if (preferredPierSide == PPS_WEST) strcpy(reply, "W"); else strcpy(reply, "B");
                quietReply = true; break;
              case '7': dtostrf(slewSpeed, 3, 1, reply); quietReply = true; break;                          // slew speed
              case '8':
#ifdef ROTATOR_ON
#ifdef MOUNT_TYPE_ALTAZM
                strcpy(reply, "D");
#else
                strcpy(reply, "R");
#endif
#else
                strcpy(reply, "N");
#endif
                quietReply = true; break; // rotator availablity 2=rotate/derotate, 1=rotate, 0=off
              default:  commandError = true;
            }
          }
          else if (parameter[0] == 'E') { // En: Get settings
            switch (parameter[1]) {
              case '0':
                // strcpy(reply, (char *)ConfighSettings().c_str());
                quietReply = true;
                break;
              case '1':
                sprintf(reply, "%ld", (long)MaxRate);
                quietReply = true;
                break;
              case '2':
                dtostrf(DegreesForAcceleration, 2, 1, reply);
                quietReply = true;
                break;
              case '3':
                sprintf(reply, "%ld", (long)round(BacklashTakeupRate));
                quietReply = true;
                break;
              case '4':
                sprintf(reply, "%ld", (long)round(StepsPerDegreeAxis1));
                quietReply = true;
                break;
              case '5':
                sprintf(reply, "%ld", (long)round(StepsPerDegreeAxis2));
                quietReply = true;
                break;
              case '6':
                dtostrf(StepsPerSecondAxis1, 3, 6, reply);
                quietReply = true;
                break;
              case '7':
                sprintf(reply, "%ld", (long)round(StepsPerWormRotationAxis1));
                quietReply = true;
                break;
              case '8':
                sprintf(reply, "%ld", (long)round(PECBufferSize));
                quietReply = true;
                break;
#ifdef MOUNT_TYPE_GEM
              case '9':
                sprintf(reply, "%ld", (long)round(minutesPastMeridianE));
                quietReply = true;
                break;
              case 'A':
                sprintf(reply, "%ld", (long)round(minutesPastMeridianW));
                quietReply = true;
                break;
#endif
              case 'B':
                sprintf(reply, "%ld", (long)round(UnderPoleLimit));
                quietReply = true;
                break;
              case 'C':
                sprintf(reply, "%ld", (long)round(MinDec));
                quietReply = true;
                break;
              case 'D':
                sprintf(reply, "%ld", (long)round(MaxDec));
                quietReply = true;
                break;
              default:
                commandError = true;
            }
          }
          else if (parameter[0] == 'F') { // Fn: Debug
            long temp;
            switch (parameter[1]) {
              case '0':
                cli();
                temp = (long)(posAxis1 - ((long)targetAxis1.part.m));
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true;
                break; // Debug0, true vs. target RA position
              case '1':
                cli();
                temp = (long)(posAxis2 - ((long)targetAxis2.part.m));
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true; break; // Debug1, true vs. target Dec position
              case '2':
                cli();
                temp = (long)trackingState;
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true; break;                   // Debug2, trackingState
              case '3':
                sprintf(reply, "%ld", (long)(az_deltaAxis1 * 1000.0 * 1.00273790935));
                quietReply = true;
                break;                   // Debug3, RA refraction tracking rate
              case '4':
                sprintf(reply, "%ld", (long)(az_deltaAxis2 * 1000.0 * 1.00273790935));
                quietReply = true;
                break;                   // Debug4, Dec refraction tracking rate
              case '5':
                sprintf(reply, "%ld", (long)(ZenithTrackingRate() * 1000.0 * 1.00273790935));
                quietReply = true;
                break;            // Debug5, Alt RA refraction tracking rate
              case '6':
                cli();
                temp = (long)(targetAxis1.part.m);
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true;
                break;            // Debug6, HA target position
              case '7':
                cli();
                temp = (long)(targetAxis2.part.m);
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true;
                break;            // Debug7, Dec target position
              case '8':
                cli();
                temp = (long)(posAxis1);
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true;
                break;                      // Debug8, HA motor position
              case '9':
                cli();
                temp = (long)(posAxis2);
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true;
                break;                      // Debug9, Dec motor position
              case 'A':
                sprintf(reply, "%ld%%", (worst_loop_time * 100L) / 9970L);
                worst_loop_time = 0;
                quietReply = true;
                break;          // DebugA, Workload
              case 'B':
                cli();
                temp = (long)(trackingTimerRateAxis1 * 1000.0);
                sei();
                sprintf(reply, "%ld", temp);
                quietReply = true;
                break; // DebugB, trackingTimerRateAxis1
              default:
                commandError = true;
            }
          }
          else
            commandError = true;
        }
        else commandError = true;
      }
      else
        //  :GZ#   Get telescope azimuth
        //         Returns: DDD*MM# or DDD*MM'SS# (based on precision setting)
        if (command[1] == 'Z')  {
          // getHor(&f, &f1);
          f1 = degRange(f1);
          if (!doubleToDms(reply, &f1, true, false))
            commandError = true;
          else
            quietReply = true;
        } else
          commandError = true;
    }
    else if (command[0] == 'M') {
      //   M - Telescope Movement Commands

      //  :MA#   Goto the target Alt and Az
      //         Returns: 0..9, see :MS#
      if (command[1] == 'A') {
        if (axis1Enabled) {
          // i = goToHor(&newTargetAlt, &newTargetAzm);
          reply[0] = i + '0'; reply[1] = 0;
          quietReply = true;
          supress_frame = true;
        } else {
          reply[0] = '4';
          reply[1] = 0;
        }
      } else if (command[1] == 'g') {
        //  :Mgdnnnn# Pulse guide command
        //          Returns: Nothing
        if ( (atoi2((char *)&parameter[1], &i)) && ((i >= 0) && (i <= 16399)) ) {
          if (((parameter[0] == 'e') || (parameter[0] == 'w')) && (guideDirAxis1 == 0)) {
#ifdef SEPARATE_PULSE_GUIDE_RATE_ON
            // startGuideAxis1(parameter[0], currentPulseGuideRate, i);
#else
            // startGuideAxis1(parameter[0], currentGuideRate, i);
#endif
            quietReply = true;
          }
          else if (((parameter[0] == 'n') || (parameter[0] == 's')) && (guideDirAxis2 == 0)) {
#ifdef SEPARATE_PULSE_GUIDE_RATE_ON
            // startGuideAxis2(parameter[0], currentPulseGuideRate, i);
#else
            // startGuideAxis2(parameter[0], currentGuideRate, i);
#endif
            quietReply = true;
          }
          else commandError = true;
        }
        else commandError = true;
      }
      // Dolphin 2021.01.09
      else if ((command[1] == 'e') || (command[1] == 'w')) {
        //  :Me# & :Mw#      Move Telescope East or West at current slew rate
        //         Returns: Nothing
        // startGuideAxis1(command[1], currentGuideRate, GUIDE_TIME_LIMIT * 1000);
          getEqu(&f0, &f1, false);
          if (command[1] == 'e') goToEqu(f0 - TRIM, f1);
          else if (command[1] == 'w') goToEqu(f0 + TRIM, f1);
        
        quietReply = true;
      }
      else if ((command[1] == 'n') || (command[1] == 's')) {
        //  :Mn# & :Ms#      Move Telescope North or South at current slew rate
        //         Returns: Nothing
        // startGuideAxis2(command[1], currentGuideRate, GUIDE_TIME_LIMIT * 1000);
          getEqu(&f0, &f1, false);
          if (command[1] == 'n') goToEqu(f0, f1 - TRIM);
          else if (command[1] == 's') goToEqu(f0, f1 + TRIM);
               
        quietReply = true;
      }
      else if (command[1] == 'P')  {
        //  :MP#   Goto the Current Position for Polar Align
        //         Returns: 0..9, see :MS#
        double r, d;
        // getEqu(&r, &d, false);
        // i = validateGoToEqu(r, d);
        if (i == 0) {
          // GeoAlign.altCor = 0.0;
          // GeoAlign.azmCor = 0.0;
          // i = goToEqu(r, d);
        }
        reply[0] = i + '0'; reply[1] = 0;
        quietReply = true;
        supress_frame = true;
      }
      else if (command[1] == 'S')  {
        //  :MS#   Goto the Target Object
        //         Returns:
        //         0=Goto is possible
        //         1=below the horizon limit
        //         2=above overhead limit
        //         3=controller in standby
        //         4=mount is parked
        //         5=Goto in progress
        //         6=outside limits (MaxDec, MinDec, UnderPoleLimit, MeridianLimit)
        //         7=hardware fault
        //         8=already in motion
        //         9=unspecified error
        i = goToEqu(newTargetRA, newTargetDec);
        reply[0] = i + '0';
        reply[1] = 0;
        quietReply = true;
        supress_frame = true;
      }
      else
        commandError = true;
    }
    else if (command[0] == 'Q') {
      //   Q - Movement Commands
      //  :Q#    Halt all slews, stops goto
      //         Returns: Nothing
      if (command[1] == 0) {
        stopMount();
        quietReply = true;
      }
      else if ((command[1] == 'e') || (command[1] == 'w')) {
        //  :Qe# & Qw#   Halt east/westward Slews
        //         Returns: Nothing
        stopGuideAxis1();
        quietReply = true;
      }
      else if ((command[1] == 'n') || (command[1] == 's')) {
        //  :Qn# & Qs#   Halt north/southward Slews
        //         Returns: Nothing
        stopGuideAxis2();
        quietReply = true;
      }
      else
        commandError = true;
    }
    else if (command[0] == 'S') {
      //   S - Telescope Set Commands

      //  :SasDD*MM#
      //         Set target object altitude to sDD*MM# or sDD*MM'SS# (based on precision setting)
      //         Returns:
      //         0 if Object is within slew range, 1 otherwise
      if (command[1] == 'a')  {
        if (!dmsToDouble(&newTargetAlt, parameter, true)) commandError = true;
      }
      else if (command[1] == 'B') {
        //  :SBn#  Set Baud Rate n for Serial-0, where n is an ASCII digit (1..9) with the following interpertation
        //         0=115.2K, 1=56.7K, 2=38.4K, 3=28.8K, 4=19.2K, 5=14.4K, 6=9600, 7=4800, 8=2400, 9=1200
        //         Returns: "1" At the current baud rate and then changes to the new rate for further communication
        i = (int)(parameter[0] - '0');
        if ((i >= 0) && (i < 10)) {
          if (process_command == COMMAND_SERIAL) {
            SerialLX200String(0, "1");
#ifdef HAL_SERIAL_TRANSMIT
            while (PSerial.transmit());
#endif
            delay(50);
            SerialLX200String(0, baudRate[i]);
            quietReply = true;
#ifdef HAL_SERIAL1_ENABLED
          }
          else if (process_command == COMMAND_SERIAL1) {
            PSerial1.print("1");
#ifdef HAL_SERIAL_TRANSMIT
            while (PSerial1.transmit());
#endif
            delay(50); PSerial1.begin(baudRate[i]);
            quietReply = true;
#endif
#ifdef HAL_SERIAL4_ENABLED
          }
          else if (process_command == COMMAND_SERIAL4) {
            PSerial4.print("1");
#ifdef HAL_SERIAL_TRANSMIT
            while (PSerial4.transmit());
#endif
            delay(50); PSerial4.begin(baudRate[i]);
            quietReply = true;
#endif
          }
          else
            commandError = true;
        }
        else
          commandError = true;
      }
      else if (command[1] == 'C')  {
        //  :SCMM/DD/YY#
        //          Change Date to MM/DD/YY
        //          Return: 0 on failure
        //                  1 on success
        if (dateToDouble(&JD, parameter)) {
          // EEPROM_writeFloat(EE_JD, JD);
          update_lst(jd2last(JD, UT1, true));
        }
        else commandError = true;
      }
      else if (command[1] == 'd')  {
        //  :SdsDD*MM#
        //          Set target object declination to sDD*MM or sDD*MM:SS depending on the current precision setting, automatically detects low/high precision
        //          Return: 0 on failure
        //                  1 on success
        if (!dmsToDouble(&newTargetDec, parameter, true)) {
          i = highPrecision; highPrecision = !highPrecision;
          if (!dmsToDouble(&newTargetDec, parameter, true)) commandError = true;
          highPrecision = i;
        }
      }
      else if (command[1] == 'g')  {
        //  :SgsDDD*MM# or :SgDDD*MM#
        //          Set current sites longitude to sDDD*MM an ASCII position string, East longitudes can be as negative or >180 degrees
        //          Return: 0 on failure
        //                  1 on success
        i = highPrecision;
        highPrecision = false;
        if ((parameter[0] == '-') || (parameter[0] == '+')) i1 = 1; else i1 = 0;
        if (!dmsToDouble(&longitude, (char *)&parameter[i1], false))
          commandError = true;
        else {
          if (parameter[0] == '-') longitude = -longitude;
          if ((longitude >= -180.0) && (longitude <= 360.0)) {
            if (longitude >= 180.0) longitude -= 360.0;
            // EEPROM_writeFloat(EE_sites + (currentSite) * 25 + 4, longitude);
          } else commandError = true;
        }
        update_lst(jd2last(JD, UT1, false));
        highPrecision = i;
      }
      else if (command[1] == 'G')  {
        //  :SGsHH#
        //  :SGsHH:MM# (where MM is 30 or 45)
        //          Set the number of hours added to local time to yield UTC
        //          Return: 0 on failure
        //                  1 on success
        if (strlen(parameter) < 7) {
          double f = 0.0;
          char *temp = strchr(parameter, ':');
          if (temp) {
            if ((temp[0] == ':') && (temp[1] == '4') && temp[2] == '5') {
              temp[0] = 0;
              f = 0.75;
            } else if ((temp[0] == ':') && (temp[1] == '3') && temp[2] == '0') {
              temp[0] = 0;
              f = 0.5;
            } else if ((temp[0] == ':') && (temp[1] == '0') && temp[2] == '0') {
              temp[0] = 0;  // force error if not :00 or :30 or :45
              f = 0.0;
            } else {
              i = -999;
            }
          }

          if ( (atoi2(parameter, &i)) && ((i >= -24) && (i <= 24))) {
            if (i < 0) timeZone = i - f; else timeZone = i + f;
            b = encodeTimeZone(timeZone) + 128;
            // EEPROM.update(EE_sites + (currentSite) * 25 + 8, b);
            update_lst(jd2last(JD, UT1, true));
          } else commandError = true;
        } else commandError = true;
      }
      else if (command[1] == 'h')  {
        //  :Sh+DD#
        //          Set the lowest elevation to which the telescope will goTo
        //          Return: 0 on failure
        //                  1 on success
        if ((parameter[0] != 0) && (strlen(parameter) < 4)) {
          if ( (atoi2(parameter, &i)) && ((i >= -30) && (i <= 30))) {
            minAlt = i;
            // EEPROM.update(EE_minAlt, minAlt + 128);
          } else commandError = true;
        } else commandError = true;
      }
      else if (command[1] == 'L')  {
        //  :SLHH:MM:SS#
        //          Set the local Time
        //          Return: 0 on failure
        //                  1 on success
        i = highPrecision; highPrecision = true;
        if (!hmsToDouble(&LMT, parameter)) commandError = true; else {
          // EEPROM_writeFloat(EE_LMT, LMT);
          UT1 = LMT + timeZone;
          update_lst(jd2last(JD, UT1, true));
        }
        highPrecision = i;
      }
      else if ((command[1] == 'M') || (command[1] == 'N') || (command[1] == 'O') || (command[1] == 'P')) {
        //  :SM<string>#
        //  :SN<string>#
        //  :SO<string>#
        //  :SP<string>#
        //          Set site name to be <string>, up to 15 characters.
        //          Return: 0 on failure
        //                  1 on success
        i = command[1] - 'M';
        // if (strlen(parameter) > 15) commandError = true; else EEPROM_writeString(EE_sites + i * 25 + 9, parameter);
      }
      else if (command[1] == 'o')  {
        //  :SoDD#
        //          Set the overhead elevation limit to DD#
        //          Return: 0 on failure
        //                  1 on success
        if ((parameter[0] != 0) && (strlen(parameter) < 3)) {
          if ( (atoi2(parameter, &i)) && ((i >= 60) && (i <= 90))) {
            maxAlt = i;
#ifdef MOUNT_TYPE_ALTAZM
            if (maxAlt > 87) maxAlt = 87;
#endif
            // EEPROM.update(EE_maxAlt, maxAlt);
          } else commandError = true;
        } else commandError = true;
      }
      else if (command[1] == 'r')  {
        //  :SrHH:MM.T#
        //  :SrHH:MM:SS#
        //          Set target object RA to HH:MM.T or HH:MM:SS based on precision setting, automatically detects low/high precision
        //          Return: 0 on failure
        //                  1 on success
        if (!hmsToDouble(&newTargetRA, parameter))
        {
          i = highPrecision;
          highPrecision = !highPrecision;
          if (!hmsToDouble(&newTargetRA, parameter))
            commandError = true;
          else
            newTargetRA *= 15.0;
          highPrecision = i;
        }
        else
          newTargetRA *= 15.0;
      }
      else if (command[1] == 'S')  {
        //  :SSHH:MM:SS#
        //          Sets the local (apparent) sideral time to HH:MM:SS
        //          Return: 0 on failure
        //                  1 on success
        i = highPrecision;
        highPrecision = true;
        if (!hmsToDouble(&f0, parameter)) commandError = true;
        else update_lst(f0);
        highPrecision = i;
      }
      else if (command[1] == 't')  {
        //  :StsDD*MM#
        //          Sets the current site latitude to sDD*MM#
        //          Return: 0 on failure
        //                  1 on success
        i = highPrecision; highPrecision = false;
        if (!dmsToDouble(&latitude, parameter, true)) {
          commandError = true;
        }
        else {
          // EEPROM_writeFloat(100 + (currentSite) * 25 + 0, latitude);
#ifdef MOUNT_TYPE_ALTAZM
          celestialPoleAxis2 = AltAzmDecStartPos;
          if (latitude < 0) celestialPoleAxis1 = 180L; else celestialPoleAxis1 = 0L;
#else
          if (latitude < 0) celestialPoleAxis2 = -90.0; else celestialPoleAxis2 = 90.0;
#endif
          cosLat = cos(latitude / Rad);
          sinLat = sin(latitude / Rad);
          if (latitude > 0.0) defaultDirAxis1 = defaultDirAxis1NCPInit; else defaultDirAxis1 = defaultDirAxis1SCPInit;
        }
        highPrecision = i;
      }
      else if (command[1] == 'T')  {
        //  :STdd.ddddd#    Set Tracking Rate
        //          Return: 0 on failure
        //                  1 on success
        if ((trackingState == TrackingSidereal) || (trackingState == TrackingNone)) {
          f0 = strtod(parameter, &conv_end);
          if ( (&parameter[0] != conv_end) && (((f0 >= 30.0) && (f0 < 90.0)) || (abs(f0) < 0.1))) {
            if (abs(f0) < 0.1) {
              trackingState = TrackingNone;
            }
            else {
              if (trackingState == TrackingNone) {
                trackingState = TrackingSidereal;
                // EnableStepperDrivers();
              }
              SetTrackingRate((f0 / 60.0) / 1.00273790935);
            }
          }
          else commandError = true;
        }
        else commandError = true;
      }
      else if (command[1] == 'X')  {
        //  :SXnn,VVVVVV...#   Set OnStep value
        //          Return: 0 on failure
        //                  1 on success
        if (parameter[2] != ',') {
          parameter[0] = 0;  // make sure command format is correct
          commandError = true;
        }
        if (parameter[0] == '0') { // 0n: Align Model
          /*
                    switch (parameter[1]) {
                      case '0': indexAxis1 = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // indexAxis1
                      case '1': indexAxis2 = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // indexAxis2
                      case '2': GeoAlign.altCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // altCor
                      case '3': GeoAlign.azmCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // azmCor
                      case '4': GeoAlign.doCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // doCor
                      case '5': GeoAlign.pdCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // pdCor
            #if defined(MOUNT_TYPE_FORK) || defined(MOUNT_TYPE_FORK_ALT)
                      case '6': GeoAlign.dfCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // ffCor
                      case '7': break;                                                               // dfCor
            #else
                      case '6': break;                                                               // ffCor
                      case '7': GeoAlign.dfCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // dfCor
            #endif
                      case '8': GeoAlign.tfCor = (double)strtol(&parameter[3], NULL, 10) / 3600.0; break; // tfCor
                      default:  commandError = true;
                    }
            // */
        }
        else if (parameter[0] == '9') { // 9n: Misc.
          switch (parameter[1]) {
            case '2': // set new acceleration rate
              maxRate = strtol(&parameter[3], NULL, 10) * 16L;
              if (maxRate < (MaxRate / 2L) * 16L) maxRate = (MaxRate / 2L) * 16L;
              if (maxRate > (MaxRate * 2L) * 16L) maxRate = (MaxRate * 2L) * 16L;
              // EEPROM_writeInt(EE_maxRate, (int)(maxRate / 16L));
              SetAccelerationRates(maxRate);
              break;
            case '3': // acceleration rate preset
              quietReply = true;
              switch (parameter[3]) {
                case '5': maxRate = MaxRate * 32L; break; // 50%
                case '4': maxRate = MaxRate * 24L; break; // 75%
                case '3': maxRate = MaxRate * 16L; break; // 100%
                case '2': maxRate = MaxRate * 12L; break; // 150%
                case '1': maxRate = MaxRate * 8L;  break; // 200%
                default:  maxRate = MaxRate * 16L;
              }
              // EEPROM_writeInt(EE_maxRate, (int)(maxRate / 16L));
              SetAccelerationRates(maxRate);
              break;
            case '5': // autoMeridianFlip
              if ((parameter[3] == '0') || (parameter[3] == '1')) {
                i = parameter[3] - '0';
                autoMeridianFlip = i;
                // EEPROM.write(EE_autoMeridianFlip, autoMeridianFlip);
              } else commandError = true;
              break;
            case '6': // preferred pier side
              switch (parameter[3]) {
                case 'E': preferredPierSide = PPS_EAST; break;
                case 'W': preferredPierSide = PPS_WEST; break;
                case 'B': preferredPierSide = PPS_BEST; break;
                default:  commandError = true;
              }
              break;
            case '7': // buzzer
              if ((parameter[3] == '0') || (parameter[3] == '1')) {
                soundEnabled = parameter[3] - '0';
              } else commandError = true;
              break;
            case '8': // pause at home on meridian flip
              if ((parameter[3] == '0') || (parameter[3] == '1')) {
                pauseHome = parameter[3] - '0';
                // EEPROM.write(EE_pauseHome, pauseHome);
              } else commandError = true;
              break;
            case '9': // continue if paused at home
              if ((parameter[3] == '1')) {
                if (waitingHome) waitingHomeContinue = true;
              } commandError = true;
              break;
            default: commandError = true;
          }
        }
        else
#ifdef MOUNT_TYPE_GEM
          if (parameter[0] == 'E') { // En: Simple value
            switch (parameter[1]) {
              case '9': // minutesPastMeridianE
                minutesPastMeridianE = (double)strtol(&parameter[3], NULL, 10);
                if (minutesPastMeridianE > 180) minutesPastMeridianE = 180;
                if (minutesPastMeridianE < -180) minutesPastMeridianE = -180;
                // EEPROM.write(EE_dpmE, round((minutesPastMeridianE * 15.0) / 60.0) + 128);
                break;
              case 'A': // minutesPastMeridianW
                minutesPastMeridianW = (double)strtol(&parameter[3], NULL, 10);
                if (minutesPastMeridianW > 180) minutesPastMeridianW = 180;
                if (minutesPastMeridianW < -180) minutesPastMeridianW = -180;
                // EEPROM.write(EE_dpmW, round((minutesPastMeridianW * 15.0) / 60.0) + 128);
                break;
              default: commandError = true;
            }
          } else
#endif
            commandError = true;
      }
      else if (command[1] == 'z')  {
        //  :SzDDD*MM#
        //          Sets the target Object Azimuth
        //          Return: 0 on failure
        //                  1 on success
        if (!dmsToDouble(&newTargetAzm, parameter, false)) commandError = true;
      }
      else
        commandError = true;
    }
    else
    {
      commandError = true;
      quietReply = 1; // Dolphin 2020.12.13
    }
    if (!quietReply) {
      if (commandError)
        reply[0] = '0';
      else
        reply[0] = '1';
      reply[1] = 0;
      supress_frame = true;
    }

    if (strlen(reply) > 0) {
      if (process_command == COMMAND_SERIAL) {
        if (cmd.checksum) checksum(reply);
        if (!supress_frame) strcat(reply, "#");
        SerialLX200String(0, reply);
      }
      if (process_command == COMMAND_SERIALX) {
        if (cmdx.checksum) checksum(reply);
        if (!supress_frame) strcat(reply, "#");
        strcpy(replyx, reply);
      }
    }
    quietReply = false;
  }
}


void stopMount() {
  if ((parkStatus == NotParked) || (parkStatus == Parking)) {
    stopGuideAxis1();
    stopGuideAxis2();
    if (trackingState == TrackingMoveTo) {
      abortSlew = true;
    }
  }
}


// calculate the checksum and add to string
void checksum(char s[]) {
  char HEXS[3] = "";
  byte cks = 0; for (unsigned int cksCount0 = 0; cksCount0 < strlen(s); cksCount0++) {
    cks += s[cksCount0];
  }
  sprintf(HEXS, "%02X", cks);
  strcat(s, HEXS);
}
