#include "Arduino.h"
#include "config.h"
#include "def.h"
#include "Serial.h"
#include "MultiWii.h"

static volatile uint8_t serialHeadRX[UART_NUMBER], serialTailRX[UART_NUMBER];
static uint8_t serialBufferRX[RX_BUFFER_SIZE][UART_NUMBER];
static volatile uint8_t serialHeadTX[UART_NUMBER], serialTailTX[UART_NUMBER];
static uint8_t serialBufferTX[TX_BUFFER_SIZE][UART_NUMBER];

static volatile uint8_t serialLX200HeadRX[UART_NUMBER], serialLX200TailRX[UART_NUMBER];
static uint8_t serialLX200BufferRX[RX_BUFFER_SIZE][UART_NUMBER];
static volatile uint8_t serialLX200HeadTX[UART_NUMBER], serialLX200TailTX[UART_NUMBER];
static uint8_t serialLX200BufferTX[TX_BUFFER_SIZE][UART_NUMBER];


// *******************************************************
// Interrupt driven UART transmitter - using a ring buffer
// *******************************************************


ISR(USART_UDRE_vect) {  // Serial 0 on a PROMINI
  uint8_t t = serialTailTX[0];
  if (serialHeadTX[0] != t) {
    if (++t >= TX_BUFFER_SIZE) t = 0;
    UDR0 = serialBufferTX[t][0];  // Transmit next byte in the ring
    serialTailTX[0] = t;
  }

  //  if (t == serialHeadTX[0]) UCSR0B &= ~(1 << UDRIE0); // Check if all data is transmitted . if yes disable transmitter UDRE interrupt
  if (t == serialHeadTX[0])
  {
    uint8_t tLX200 = serialLX200TailTX[0];
    if (serialLX200HeadTX[0] != tLX200) {
      if (++tLX200 >= TX_BUFFER_SIZE) tLX200 = 0;
      UDR0 = serialLX200BufferTX[tLX200][0];  // Transmit next byte in the ring
      serialLX200TailTX[0] = tLX200;
    }

    if (tLX200 == serialLX200HeadTX[0]) UCSR0B &= ~(1 << UDRIE0); // Check if all data is transmitted . if yes disable transmitter UDRE interrupt
  }
}

void UartSendData(uint8_t port) {
  UCSR0B |= (1 << UDRIE0);
}

void SerialOpen(uint8_t port, uint32_t baud) {
  uint8_t h = ((F_CPU  / 4 / baud - 1) / 2) >> 8;
  uint8_t l = ((F_CPU  / 4 / baud - 1) / 2);
  switch (port) {
    case 0: UCSR0A  = (1 << U2X0); UBRR0H = h; UBRR0L = l; UCSR0B |= (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0); break;
  }
}

void SerialEnd(uint8_t port) {
  switch (port) {
    case 0: UCSR0B &= ~((1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0) | (1 << UDRIE0)); break;
  }
}

// we don't care about ring buffer overflow (head->tail) to avoid a test condition : data is lost anyway if it happens
void store_uart_in_buf(uint8_t data, uint8_t portnum) {
  uint8_t h = serialHeadRX[portnum];
  serialBufferRX[h++][portnum] = data;
  if (h >= RX_BUFFER_SIZE) h = 0;
  serialHeadRX[portnum] = h;

  uint8_t hLX200 = serialLX200HeadRX[portnum];
  serialLX200BufferRX[hLX200++][portnum] = data;
  if (hLX200 >= RX_BUFFER_SIZE) hLX200 = 0;
  serialLX200HeadRX[portnum] = hLX200;
}

ISR(USART_RX_vect)  {
  store_uart_in_buf(UDR0, 0);
}

uint8_t SerialRead(uint8_t port) {
  uint8_t t = serialTailRX[port];
  uint8_t c = serialBufferRX[t][port];
  if (serialHeadRX[port] != t) {
    if (++t >= RX_BUFFER_SIZE) t = 0;
    serialTailRX[port] = t;
  }
  return c;
}

uint8_t SerialLX200Read(uint8_t port) {
  uint8_t tLX200 = serialLX200TailRX[port];
  uint8_t cLX200 = serialLX200BufferRX[tLX200][port];
  if (serialLX200HeadRX[port] != tLX200) {
    if (++tLX200 >= RX_BUFFER_SIZE) tLX200 = 0;
    serialLX200TailRX[port] = tLX200;
  }
  return cLX200;
}

uint8_t SerialAvailable(uint8_t port) {
  return ((uint8_t)(serialHeadRX[port] - serialTailRX[port])) % RX_BUFFER_SIZE;
}

uint8_t SerialLX200Available(uint8_t port) {
  return ((uint8_t)(serialLX200HeadRX[port] - serialLX200TailRX[port])) % RX_BUFFER_SIZE;
}

uint8_t SerialUsedTXBuff(uint8_t port) {
  return ((uint8_t)(serialHeadTX[port] - serialTailTX[port])) % TX_BUFFER_SIZE;
}

uint8_t SerialLX200UsedTXBuff(uint8_t port) {
  return ((uint8_t)(serialLX200HeadTX[port] - serialLX200TailTX[port])) % TX_BUFFER_SIZE;
}

void SerialSerialize(uint8_t port, uint8_t a) {
  uint8_t t = serialHeadTX[port];
  if (++t >= TX_BUFFER_SIZE) t = 0;
  serialBufferTX[t][port] = a;
  serialHeadTX[port] = t;
}

void SerialLX200Serialize(uint8_t port, uint8_t a) {
  uint8_t tLX200 = serialLX200HeadTX[port];
  if (++tLX200 >= TX_BUFFER_SIZE) tLX200 = 0;
  serialLX200BufferTX[tLX200][port] = a;
  serialLX200HeadTX[port] = tLX200;
}

void SerialWrite(uint8_t port, uint8_t c) {
  SerialSerialize(port, c); UartSendData(port);
}

void SerialLX200Write(uint8_t port, uint8_t c) {
  SerialLX200Serialize(port, c); UartSendData(port);
}

void SerialLX200String(uint8_t port, uint8_t *c)
{
  uint8_t nLen = strlen((char *)c);

  for (uint8_t i = 0; i < nLen; i++)
  {
    SerialLX200Serialize(port, c[i]);
  }
  UartSendData(port);
}
