#include "Arduino.h"

#include "EQ.h"
#include "TMC2208.h"

// Note: _stepdelay has to be an unsigned long to be able to avoid overflow of the timers thanks to register arithmetic
TMC2208::TMC2208(int _pDir, int _pStep, int _pEnable, unsigned long _stepdelay, int _nRev)
{
  pDir = _pDir;
  pStep = _pStep;
  pEnable = _pEnable;
  nRev = _nRev;
  if (pDir >= 0)
  {
    pinMode(pDir, OUTPUT);
  }
  if (pStep >= 0)
  {
    pinMode(pStep, OUTPUT);
  }
  if (pEnable >= 0)
  {
    pinMode(pEnable, OUTPUT);
  }

  nState = STATE_DISABLE;
  steptime = _stepdelay;
  stepstogo = 0;
  _lastStepTime = micros();
}

// run() should be computationally cheap if false. Just check if we have to move
// (maybe checking objective vs current step? Or with a flag?) and send to the
// function. Or else return quickly so that we don't stall the loop.
boolean TMC2208::run()
{
  // Do time check first
  unsigned long time = micros();

  if (nState == STATE_DISABLE)
  {
    MyDebug2();

    if (stepstogo > 0)
    {
      if (pEnable >= 0)
      {
        digitalWrite(pEnable, LOW);
      }
      if (pDir >= 0)
      {
        digitalWrite(pDir, HIGH);
      }
      if (pStep >= 0)
      {
        digitalWrite(pStep, HIGH);
        digitalWrite(13, HIGH);
      }

      stepstogo--;
      _lastStepTime = micros();
      nState = STATE_STEP_HIGH;
    }
    else if (stepstogo < 0)
    {
      if (pEnable >= 0)
      {
        digitalWrite(pEnable, LOW);
      }
      if (pDir >= 0)
      {
        digitalWrite(pDir, LOW);
      }
      if (pStep >= 0)
      {
        digitalWrite(pStep, HIGH);
        digitalWrite(13, HIGH);
      }

      stepstogo++;
      _lastStepTime = micros();
      nState = STATE_STEP_HIGH;
    }
  }
  else if (nState == STATE_STEP_HIGH)
  {

    MyDebug();

    if (time - _lastStepTime >= steptime)
    {
      if (pStep >= 0)
      {
        digitalWrite(pStep, LOW);
        digitalWrite(13, LOW);
      }

      _lastStepTime = micros();
      nState = STATE_STEP_LOW;
    }
  }
  else if (nState == STATE_STEP_LOW)
  {

    MyDebug();

    if (time - _lastStepTime >= steptime)
    {
      if (stepstogo > 0)
      {
        if (pDir >= 0)
        {
          digitalWrite(pDir, HIGH);
        }
        if (pStep >= 0)
        {
          digitalWrite(pStep, HIGH);
          digitalWrite(13, HIGH);
        }

        stepstogo--;
        _lastStepTime = micros();
        nState = STATE_STEP_HIGH;
      }
      else if (stepstogo < 0)
      {
        if (pDir >= 0)
        {
          digitalWrite(pDir, LOW);
        }
        if (pStep >= 0)
        {
          digitalWrite(pStep, HIGH);
          digitalWrite(13, HIGH);
        }

        stepstogo++;
        _lastStepTime = micros();
        nState = STATE_STEP_HIGH;
      }
      else
      {
        _lastStepTime = micros();
        nState = STATE_ENABLE;
      }
    }
  }
  else if (nState == STATE_ENABLE)
  {
    if (stepstogo > 0)
    {
      if (pDir >= 0)
      {
        digitalWrite(pDir, HIGH);
      }
      if (pStep >= 0)
      {
        digitalWrite(pStep, HIGH);
        digitalWrite(13, HIGH);
      }

      stepstogo--;
      _lastStepTime = micros();
      nState = STATE_STEP_HIGH;
    }
    else if (stepstogo < 0)
    {
      if (pDir >= 0)
      {
        digitalWrite(pDir, LOW);
      }
      if (pStep >= 0)
      {
        digitalWrite(pStep, HIGH);
        digitalWrite(13, HIGH);
      }

      stepstogo++;
      _lastStepTime = micros();
      nState = STATE_STEP_HIGH;
    }
    else
    {
      if (time - _lastStepTime >= STOP_DELAY_TIME)
      {
        if (pEnable >= 0)
        {
          digitalWrite(pEnable, HIGH);
        }
        if (pDir >= 0)
        {
          digitalWrite(pDir, LOW);
        }
        if (pStep >= 0)
        {
          digitalWrite(pStep, LOW);
          digitalWrite(13, LOW);
        }

        _lastStepTime = micros();

        nState = STATE_DISABLE;
      }
    }
  }
  else
  {
    nState = STATE_DISABLE;
  }

  return true;
}

// Setup a movement. Set stepstogo.
void TMC2208::move(long steps) {
  if (nRev == 0)
  {
    stepstogo += steps;
  }
  else
  {
    stepstogo -= steps;
  }
  run();
}

// Returns steps to go
int TMC2208::stepsToGo()
{
  return stepstogo;
}

// Pulls pins low to save power and avoid heat build up when not moving.
void TMC2208::stop()
{
  stepstogo = 0;
  run();
}
